@@ -47,7 +47,7 @@
  *
  * Since: 1.6
  */
-#define QMI_MINOR_VERSION (24)
+#define QMI_MINOR_VERSION (26)
 
 /**
  * QMI_MICRO_VERSION:
@@ -57,7 +57,7 @@
  *
  * Since: 1.6
  */
-#define QMI_MICRO_VERSION (12)
+#define QMI_MICRO_VERSION (2)
 
 /**
  * QMI_CHECK_VERSION:
@@ -94,4 +94,21 @@
  */
 #define QMI_MBIM_QMUX_SUPPORTED 1
 
+/**
+ * QMI_QRTR_SUPPORTED:
+ *
+ * Symbol to expose wether QMI over QRTR is supported. The symbol is always
+ * defined and set to either or 1 or 0.
+ *
+ * E.g.:
+ * |[
+ *  #if QMI_QRTR_SUPPORTED
+ *      // do something
+ *  #endif
+ * ]|
+ *
+ * Since: 1.26
+ */
+#define QMI_QRTR_SUPPORTED 1
+
 #endif /* _QMI_VERSION_H_ */
