@@ -33,11 +33,18 @@
 #include "qmicli.h"
 #include "qmicli-helpers.h"
 
+#if defined HAVE_QMI_SERVICE_UIM
+
 /* Context */
 typedef struct {
     QmiDevice *device;
     QmiClientUim *client;
     GCancellable *cancellable;
+
+    /* For Slot Status indication */
+    guint slot_status_indication_id;
+    gboolean warned_slot_ext_info_unavailable;
+    gboolean warned_slot_eid_info_unavailable;
 } Context;
 static Context *ctx;
 
@@ -52,64 +59,114 @@ static gchar *get_file_attributes_str;
 static gchar *sim_power_on_str;
 static gchar *sim_power_off_str;
 static gchar *change_provisioning_session_str;
+static gchar *switch_slot_str;
 static gboolean get_card_status_flag;
 static gboolean get_supported_messages_flag;
+static gboolean get_slot_status_flag;
+static gboolean monitor_slot_status_flag;
 static gboolean reset_flag;
 static gboolean noop_flag;
 
+#undef VALIDATE_UNKNOWN
+#define VALIDATE_UNKNOWN(str) (str ? str : "unknown")
+
 static GOptionEntry entries[] = {
+#if defined HAVE_QMI_MESSAGE_UIM_SET_PIN_PROTECTION
     { "uim-set-pin-protection", 0, 0, G_OPTION_ARG_STRING, &set_pin_protection_str,
       "Set PIN protection",
       "[(PIN1|PIN2|UPIN),(disable|enable),(current PIN)]"
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_VERIFY_PIN
     { "uim-verify-pin", 0, 0, G_OPTION_ARG_STRING, &verify_pin_str,
       "Verify PIN",
       "[(PIN1|PIN2|UPIN),(current PIN)]",
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_UNBLOCK_PIN
     { "uim-unblock-pin", 0, 0, G_OPTION_ARG_STRING, &unblock_pin_str,
       "Unblock PIN",
       "[(PIN1|PIN2|UPIN),(PUK),(new PIN)]",
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_CHANGE_PIN
     { "uim-change-pin", 0, 0, G_OPTION_ARG_STRING, &change_pin_str,
       "Change PIN",
       "[(PIN1|PIN2|UPIN),(old PIN),(new PIN)]",
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT
     { "uim-read-transparent", 0, 0, G_OPTION_ARG_STRING, &read_transparent_str,
       "Read a transparent file given the file path",
       "[0xNNNN,0xNNNN,...]"
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES
     { "uim-get-file-attributes", 0, 0, G_OPTION_ARG_STRING, &get_file_attributes_str,
       "Get the attributes of a given file",
       "[0xNNNN,0xNNNN,...]"
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_READ_RECORD
     { "uim-read-record", 0, 0, G_OPTION_ARG_STRING, &read_record_str,
       "Read a record from given file (allowed keys: record-number, record-length, file ([0xNNNN-0xNNNN,...])",
       "[\"key=value,...\"]"
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_GET_CARD_STATUS
     { "uim-get-card-status", 0, 0, G_OPTION_ARG_NONE, &get_card_status_flag,
       "Get card status",
       NULL
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_GET_SUPPORTED_MESSAGES
     { "uim-get-supported-messages", 0, 0, G_OPTION_ARG_NONE, &get_supported_messages_flag,
       "Get supported messages",
       NULL
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_POWER_ON_SIM
     { "uim-sim-power-on", 0, 0, G_OPTION_ARG_STRING, &sim_power_on_str,
       "Power on SIM card",
       "[(slot number)]"
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_POWER_OFF_SIM
     { "uim-sim-power-off", 0, 0, G_OPTION_ARG_STRING, &sim_power_off_str,
       "Power off SIM card",
       "[(slot number)]"
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_CHANGE_PROVISIONING_SESSION
     { "uim-change-provisioning-session", 0, 0, G_OPTION_ARG_STRING, &change_provisioning_session_str,
       "Change provisioning session (allowed keys: session-type, activate, slot, aid)",
       "[\"key=value,...\"]"
     },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS
+    { "uim-get-slot-status", 0, 0, G_OPTION_ARG_NONE, &get_slot_status_flag,
+      "Get slot status",
+      NULL
+    },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_SWITCH_SLOT && defined HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS
+    { "uim-switch-slot", 0, 0, G_OPTION_ARG_STRING, &switch_slot_str,
+      "Switch active physical slot",
+      "[(slot number)]"
+    },
+#endif
+#if defined HAVE_QMI_INDICATION_UIM_SLOT_STATUS
+    { "uim-monitor-slot-status", 0, 0, G_OPTION_ARG_NONE, &monitor_slot_status_flag,
+      "Watch for slot status indications",
+      NULL
+    },
+#endif
+#if defined HAVE_QMI_MESSAGE_UIM_RESET
     { "uim-reset", 0, 0, G_OPTION_ARG_NONE, &reset_flag,
       "Reset the service state",
       NULL
     },
+#endif
     { "uim-noop", 0, 0, G_OPTION_ARG_NONE, &noop_flag,
       "Just allocate or release a UIM client. Use with `--client-no-release-cid' and/or `--client-cid'",
       NULL
@@ -123,7 +180,7 @@ qmicli_uim_get_option_group (void)
     GOptionGroup *group;
 
     group = g_option_group_new ("uim",
-                                "UIM options",
+                                "UIM options:",
                                 "Show User Identity Module options",
                                 NULL,
                                 NULL);
@@ -151,8 +208,11 @@ qmicli_uim_options_enabled (void)
                  !!sim_power_on_str +
                  !!sim_power_off_str +
                  !!change_provisioning_session_str +
+                 !!switch_slot_str +
                  get_card_status_flag +
                  get_supported_messages_flag +
+                 get_slot_status_flag +
+                 monitor_slot_status_flag +
                  reset_flag +
                  noop_flag);
 
@@ -161,6 +221,9 @@ qmicli_uim_options_enabled (void)
         exit (EXIT_FAILURE);
     }
 
+    if (monitor_slot_status_flag)
+        qmicli_expect_indications ();
+
     checked = TRUE;
     return !!n_actions;
 }
@@ -171,6 +234,10 @@ context_free (Context *context)
     if (!context)
         return;
 
+    if (context->slot_status_indication_id)
+        g_signal_handler_disconnect (context->client,
+                                     context->slot_status_indication_id);
+
     if (context->client)
         g_object_unref (context->client);
     g_object_unref (context->cancellable);
@@ -186,6 +253,8 @@ operation_shutdown (gboolean operation_status)
     qmicli_async_operation_done (operation_status, FALSE);
 }
 
+#if defined HAVE_QMI_MESSAGE_UIM_SET_PIN_PROTECTION
+
 static QmiMessageUimSetPinProtectionInput *
 set_pin_protection_input_create (const gchar *str)
 {
@@ -280,6 +349,10 @@ set_pin_protection_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_SET_PIN_PROTECTION */
+
+#if defined HAVE_QMI_MESSAGE_UIM_VERIFY_PIN
+
 static QmiMessageUimVerifyPinInput *
 verify_pin_input_create (const gchar *str)
 {
@@ -371,6 +444,10 @@ verify_pin_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_VERIFY_PIN */
+
+#if defined HAVE_QMI_MESSAGE_UIM_UNBLOCK_PIN
+
 static QmiMessageUimUnblockPinInput *
 unblock_pin_input_create (const gchar *str)
 {
@@ -465,6 +542,10 @@ unblock_pin_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_UNBLOCK_PIN */
+
+#if defined HAVE_QMI_MESSAGE_UIM_CHANGE_PIN
+
 static QmiMessageUimChangePinInput *
 change_pin_input_create (const gchar *str)
 {
@@ -559,6 +640,10 @@ change_pin_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_CHANGE_PIN */
+
+#if defined HAVE_QMI_MESSAGE_UIM_GET_SUPPORTED_MESSAGES
+
 static void
 get_supported_messages_ready (QmiClientUim *client,
                               GAsyncResult *res)
@@ -597,6 +682,10 @@ get_supported_messages_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_GET_SUPPORTED_MESSAGES */
+
+#if defined HAVE_QMI_MESSAGE_UIM_POWER_ON_SIM
+
 static QmiMessageUimPowerOnSimInput *
 power_on_sim_input_create (const gchar *slot_str)
 {
@@ -606,7 +695,7 @@ power_on_sim_input_create (const gchar *slot_str)
 
     input = qmi_message_uim_power_on_sim_input_new ();
 
-    if (!qmicli_read_uint_from_string (sim_power_on_str, &slot) || (slot > G_MAXUINT8)) {
+    if (!qmicli_read_uint_from_string (slot_str, &slot) || (slot > G_MAXUINT8)) {
         g_printerr ("error: invalid slot number\n");
         return NULL;
     }
@@ -651,6 +740,10 @@ power_on_sim_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_POWER_ON_SIM */
+
+#if defined HAVE_QMI_MESSAGE_UIM_POWER_OFF_SIM
+
 static QmiMessageUimPowerOffSimInput *
 power_off_sim_input_create (const gchar *slot_str)
 {
@@ -660,7 +753,7 @@ power_off_sim_input_create (const gchar *slot_str)
 
     input = qmi_message_uim_power_off_sim_input_new ();
 
-    if (!qmicli_read_uint_from_string (sim_power_off_str, &slot) || (slot > G_MAXUINT8)) {
+    if (!qmicli_read_uint_from_string (slot_str, &slot) || (slot > G_MAXUINT8)) {
         g_printerr ("error: invalid slot number\n");
         return NULL;
     }
@@ -705,6 +798,10 @@ power_off_sim_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_POWER_OFF_SIM */
+
+#if defined HAVE_QMI_MESSAGE_UIM_CHANGE_PROVISIONING_SESSION
+
 typedef struct {
     QmiUimSessionType  session_type;
     gboolean           session_type_set;
@@ -853,6 +950,419 @@ change_provisioning_session_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_CHANGE_PROVISIONING_SESSION */
+
+#if (defined HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS || \
+     defined HAVE_QMI_INDICATION_UIM_SLOT_STATUS)
+
+static const gchar bcd_chars[] = "0123456789\0\0\0\0\0\0";
+
+static gchar *
+decode_iccid (const gchar *bcd, gsize bcd_len)
+{
+    GString *str;
+    gsize i;
+
+    if (!bcd)
+        return NULL;
+
+    str = g_string_sized_new (bcd_len * 2 + 1);
+    for (i = 0; i < bcd_len; i++) {
+        str = g_string_append_c (str, bcd_chars[bcd[i] & 0xF]);
+        str = g_string_append_c (str, bcd_chars[(bcd[i] >> 4) & 0xF]);
+    }
+    return g_string_free (str, FALSE);
+}
+
+#define EID_LENGTH 16
+
+static gchar *
+decode_eid (const gchar *eid, gsize eid_len)
+{
+    GString *str;
+    gsize i;
+
+    if (!eid)
+        return NULL;
+    if (eid_len != EID_LENGTH)
+        return NULL;
+
+    str = g_string_sized_new (eid_len * 2 + 1);
+    for (i = 0; i < eid_len; i++) {
+        str = g_string_append_c (str, bcd_chars[(eid[i] >> 4) & 0xF]);
+        str = g_string_append_c (str, bcd_chars[eid[i] & 0xF]);
+    }
+    return g_string_free (str, FALSE);
+}
+
+static void
+print_slot_status (GArray *physical_slots,
+                   GArray *ext_information,
+                   GArray *slot_eids)
+{
+    guint i;
+
+    if (ext_information && physical_slots->len != ext_information->len) {
+        g_print ("Malformed extended information data");
+        ext_information = NULL;
+    }
+
+    if (slot_eids && physical_slots->len != slot_eids->len) {
+        g_print ("Malformed slot EID data");
+        slot_eids = NULL;
+    }
+
+    for (i = 0; i < physical_slots->len; i++) {
+        QmiPhysicalSlotStatusSlot *slot_status;
+        QmiPhysicalSlotInformationSlot *slot_info = NULL;
+        GArray *slot_eid = NULL;
+        g_autofree gchar *iccid = NULL;
+        g_autofree gchar *eid = NULL;
+
+        slot_status = &g_array_index (physical_slots, QmiPhysicalSlotStatusSlot, i);
+
+        g_print ("  Physical slot %u:\n", i + 1);
+        g_print ("     Card status: %s\n",
+                 qmi_uim_physical_card_state_get_string (slot_status->physical_card_status));
+        g_print ("     Slot status: %s\n",
+                 qmi_uim_slot_state_get_string (slot_status->physical_slot_status));
+
+        if (slot_status->physical_slot_status == QMI_UIM_SLOT_STATE_ACTIVE)
+            g_print ("    Logical slot: %u\n", slot_status->logical_slot);
+
+        if (slot_status->physical_card_status != QMI_UIM_PHYSICAL_CARD_STATE_PRESENT)
+            continue;
+
+        if (slot_status->iccid->len)
+            iccid = decode_iccid (slot_status->iccid->data, slot_status->iccid->len);
+        g_print ("           ICCID: %s\n", VALIDATE_UNKNOWN (iccid));
+
+        /* Extended information, if available */
+        if (!ext_information)
+            continue;
+
+        slot_info = &g_array_index (ext_information, QmiPhysicalSlotInformationSlot, i);
+        g_print ("        Protocol: %s\n",
+                 qmi_uim_card_protocol_get_string (slot_info->card_protocol));
+        g_print ("        Num apps: %u\n", slot_info->valid_applications);
+        g_print ("        Is eUICC: %s\n", slot_info->is_euicc ? "yes" : "no");
+
+        /* EID info, if available and this is an eUICC */
+        if (!slot_info->is_euicc || !slot_eids)
+            continue;
+
+        slot_eid = g_array_index (slot_eids, GArray *, i);
+        if (slot_eid->len)
+            eid = decode_eid (slot_eid->data, slot_eid->len);
+        g_print ("             EID: %s\n", VALIDATE_UNKNOWN (eid));
+    }
+}
+
+#endif
+
+#if defined HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS
+
+static void
+get_slot_status_ready (QmiClientUim *client,
+                       GAsyncResult *res)
+{
+    QmiMessageUimGetSlotStatusOutput *output;
+    GArray *physical_slots;
+    GArray *ext_information = NULL;
+    GArray *slot_eids = NULL;
+    GError *error = NULL;
+
+    output = qmi_client_uim_get_slot_status_finish (client, res, &error);
+    if (!output) {
+        g_printerr ("error: operation failed: %s\n", error->message);
+        g_error_free (error);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    if (!qmi_message_uim_get_slot_status_output_get_result (output, &error)) {
+        g_printerr ("error: could not get slots status: %s\n", error->message);
+        g_error_free (error);
+        qmi_message_uim_get_slot_status_output_unref (output);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    g_print ("[%s] Successfully got slots status\n",
+             qmi_device_get_path_display (ctx->device));
+
+    if (!qmi_message_uim_get_slot_status_output_get_physical_slot_status (
+            output, &physical_slots, &error)) {
+        g_printerr ("error: could not parse slots status response: %s\n", error->message);
+        g_error_free (error);
+        qmi_message_uim_get_slot_status_output_unref (output);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    if (!qmi_message_uim_get_slot_status_output_get_physical_slot_information (
+            output, &ext_information, &error)) {
+        /* Recoverable, just print less information per slot */
+        g_print ("  Extended slots information is unavailable: %s\n", error->message);
+        g_clear_error (&error);
+    }
+
+    if (!qmi_message_uim_get_slot_status_output_get_slot_eid_information (
+          output, &slot_eids, &error)) {
+        /* Recoverable, just print less information per slot */
+        g_print ("  Slot EID information is unavailable: %s\n", error->message);
+        g_clear_error (&error);
+    }
+
+    g_print ("[%s] %u physical slots found:\n",
+             qmi_device_get_path_display (ctx->device), physical_slots->len);
+
+    print_slot_status (physical_slots, ext_information, slot_eids);
+
+    qmi_message_uim_get_slot_status_output_unref (output);
+    operation_shutdown (TRUE);
+}
+
+#endif /* HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS */
+
+#if defined HAVE_QMI_MESSAGE_UIM_SWITCH_SLOT && defined HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS
+
+static void
+switch_slot_ready (QmiClientUim *client,
+                   GAsyncResult *res)
+{
+    QmiMessageUimSwitchSlotOutput *output;
+    GError *error = NULL;
+
+    output = qmi_client_uim_switch_slot_finish (client, res, &error);
+    if (!output) {
+        g_printerr ("error: operation failed: %s\n", error->message);
+        g_error_free (error);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    if (!qmi_message_uim_switch_slot_output_get_result (output, &error)) {
+        g_printerr ("error: couldn't switch slots: %s\n", error->message);
+        g_error_free (error);
+        qmi_message_uim_switch_slot_output_unref (output);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    g_print ("[%s] Successfully switched slots\n",
+             qmi_device_get_path_display (ctx->device));
+
+    qmi_message_uim_switch_slot_output_unref (output);
+    operation_shutdown (TRUE);
+}
+
+static QmiMessageUimSwitchSlotInput *
+switch_slot_input_create (guint logical_slot,
+                          guint physical_slot)
+{
+    QmiMessageUimSwitchSlotInput *input;
+    GError                       *error = NULL;
+
+    input = qmi_message_uim_switch_slot_input_new ();
+
+    if (!qmi_message_uim_switch_slot_input_set_logical_slot (input, logical_slot, &error) ||
+        !qmi_message_uim_switch_slot_input_set_physical_slot (input, physical_slot, &error)) {
+        g_printerr ("error: could not create switch slot input: %s\n", error->message);
+        g_error_free (error);
+        g_clear_pointer (&input, qmi_message_uim_switch_slot_input_unref);
+    }
+
+    return input;
+}
+
+static void
+get_active_logical_slot_ready (QmiClientUim *client,
+                               GAsyncResult *res,
+                               gpointer user_data)
+{
+    QmiMessageUimGetSlotStatusOutput *output;
+    QmiMessageUimSwitchSlotInput *input;
+    GArray *physical_slots;
+    guint physical_slot_id;
+    guint active_logical_slot_id = 0;
+    guint i;
+    GError *error = NULL;
+
+    physical_slot_id = GPOINTER_TO_UINT (user_data);
+
+    output = qmi_client_uim_get_slot_status_finish (client, res, &error);
+    if (!output) {
+        g_printerr ("error: operation failed: %s\n", error->message);
+        g_error_free (error);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    if (!qmi_message_uim_get_slot_status_output_get_result (output, &error)) {
+        g_printerr ("error: could not get slots status: %s\n", error->message);
+        g_error_free (error);
+        qmi_message_uim_get_slot_status_output_unref (output);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    if (!qmi_message_uim_get_slot_status_output_get_physical_slot_status (
+          output, &physical_slots, &error)) {
+        g_printerr ("error: could not parse slots status response: %s\n", error->message);
+        g_error_free (error);
+        qmi_message_uim_get_slot_status_output_unref (output);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    /* Ensure the physical slot is available. */
+    if (physical_slot_id > physical_slots->len) {
+        g_printerr ("error: physical slot %u is unavailable\n", physical_slot_id);
+        qmi_message_uim_get_slot_status_output_unref (output);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    /* Find active logical slot */
+    for (i = 0; i < physical_slots->len; i++) {
+        QmiPhysicalSlotStatusSlot *slot_status;
+
+        slot_status = &g_array_index (physical_slots, QmiPhysicalSlotStatusSlot, i);
+        if (slot_status->physical_slot_status == QMI_UIM_SLOT_STATE_ACTIVE) {
+            active_logical_slot_id = slot_status->logical_slot;
+            break;
+        }
+    }
+    qmi_message_uim_get_slot_status_output_unref (output);
+
+    if (active_logical_slot_id == 0) {
+        g_printerr ("error: no active logical slot\n");
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    input = switch_slot_input_create (active_logical_slot_id, physical_slot_id);
+    qmi_client_uim_switch_slot (ctx->client,
+                                input,
+                                10,
+                                ctx->cancellable,
+                                (GAsyncReadyCallback)switch_slot_ready,
+                                ctx);
+    qmi_message_uim_switch_slot_input_unref (input);
+}
+
+#endif /* HAVE_QMI_MESSAGE_UIM_SWITCH_SLOT && HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS */
+
+#if defined HAVE_QMI_INDICATION_UIM_SLOT_STATUS
+
+static void
+monitoring_cancelled (GCancellable *cancellable)
+{
+    operation_shutdown (TRUE);
+}
+
+static void
+slot_status_received (QmiClientUim *client,
+                      QmiIndicationUimSlotStatusOutput *output)
+{
+    GArray *physical_slots;
+    GArray *ext_information = NULL;
+    GArray *slot_eids = NULL;
+    GError *error = NULL;
+
+    g_print ("[%s] Received slot status indication:\n",
+             qmi_device_get_path_display (ctx->device));
+
+    if (!qmi_indication_uim_slot_status_output_get_physical_slot_status (
+          output, &physical_slots, &error)) {
+        g_printerr ("error: could not parse slots status: %s\n", error->message);
+        g_error_free (error);
+        qmi_indication_uim_slot_status_output_unref (output);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    /* Both of these are recoverable, just print less information per slot. Only print
+     * these messages once rather than every time we get an indication */
+    if (!qmi_indication_uim_slot_status_output_get_physical_slot_information (
+          output, &ext_information, &error) && !ctx->warned_slot_ext_info_unavailable) {
+        g_print ("  Extended slots information is unavailable: %s\n", error->message);
+        ctx->warned_slot_ext_info_unavailable = TRUE;
+        g_clear_error (&error);
+    }
+
+    if (!qmi_indication_uim_slot_status_output_get_slot_eid_information (
+          output, &slot_eids, &error) && !ctx->warned_slot_eid_info_unavailable) {
+        g_print ("  Slot EID information is unavailable: %s\n", error->message);
+        ctx->warned_slot_eid_info_unavailable = TRUE;
+        g_clear_error (&error);
+    }
+
+    print_slot_status (physical_slots, ext_information, slot_eids);
+
+    qmi_indication_uim_slot_status_output_unref (output);
+}
+
+static void
+register_physical_slot_status_events_ready (QmiClientUim *client,
+                                            GAsyncResult *res)
+{
+    QmiMessageUimRegisterEventsOutput *output;
+    GError                            *error = NULL;
+
+    output = qmi_client_uim_register_events_finish (client, res, &error);
+    if (!output) {
+        g_printerr ("error: operation failed: %s\n", error->message);
+        g_error_free (error);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    if (!qmi_message_uim_register_events_output_get_result (output, &error)) {
+        g_printerr ("error: could not register slot status change events: %s\n", error->message);
+        qmi_message_uim_register_events_output_unref (output);
+        g_error_free (error);
+        operation_shutdown (FALSE);
+        return;
+    }
+
+    g_debug ("Registered physical slot status change events...");
+    ctx->slot_status_indication_id =
+        g_signal_connect (ctx->client,
+                          "slot-status",
+                          G_CALLBACK (slot_status_received),
+                          NULL);
+
+    /* User can use Ctrl+C to cancel the monitoring at any time */
+    g_cancellable_connect (ctx->cancellable,
+                           G_CALLBACK (monitoring_cancelled),
+                           NULL,
+                           NULL);
+}
+
+static void
+register_physical_slot_status_events (void)
+{
+    QmiMessageUimRegisterEventsInput *re_input;
+
+    re_input = qmi_message_uim_register_events_input_new ();
+    qmi_message_uim_register_events_input_set_event_registration_mask (
+        re_input, QMI_UIM_EVENT_REGISTRATION_FLAG_PHYSICAL_SLOT_STATUS, NULL);
+    qmi_client_uim_register_events (
+        ctx->client,
+        re_input,
+        10,
+        ctx->cancellable,
+        (GAsyncReadyCallback) register_physical_slot_status_events_ready,
+        NULL);
+    qmi_message_uim_register_events_input_unref (re_input);
+}
+
+#endif /* HAVE_QMI_INDICATION_UIM_SLOT_STATUS */
+
+#if defined HAVE_QMI_MESSAGE_UIM_RESET
+
 static void
 reset_ready (QmiClientUim *client,
              GAsyncResult *res)
@@ -883,6 +1393,8 @@ reset_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif
+
 static gboolean
 noop_cb (gpointer unused)
 {
@@ -890,6 +1402,8 @@ noop_cb (gpointer unused)
     return FALSE;
 }
 
+#if defined HAVE_QMI_MESSAGE_UIM_GET_CARD_STATUS
+
 static void
 get_card_status_ready (QmiClientUim *client,
                        GAsyncResult *res)
@@ -931,9 +1445,6 @@ get_card_status_ready (QmiClientUim *client,
         &cards,
         NULL);
 
-#undef VALIDATE_UNKNOWN
-#define VALIDATE_UNKNOWN(str) (str ? str : "unknown")
-
     g_print ("Provisioning applications:\n");
     if (index_gw_primary == 0xFFFF)
         g_print ("\tPrimary GW:   session doesn't exist\n");
@@ -975,7 +1486,7 @@ get_card_status_ready (QmiClientUim *client,
             g_print ("\tCard state: '%s'\n",
                      qmi_uim_card_state_get_string (card->card_state));
         else
-            g_print ("\tCard state: '%s: %s (%u)\n",
+            g_print ("\tCard state: '%s: %s (%u)'\n",
                      qmi_uim_card_state_get_string (card->card_state),
                      VALIDATE_UNKNOWN (qmi_uim_card_error_get_string (card->error_code)),
                      card->error_code);
@@ -1040,6 +1551,12 @@ get_card_status_ready (QmiClientUim *client,
     operation_shutdown (TRUE);
 }
 
+#endif /* defined HAVE_QMI_MESSAGE_UIM_GET_CARD_STATUS */
+
+#if defined HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT || \
+    defined HAVE_QMI_MESSAGE_UIM_READ_RECORD || \
+    defined HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES
+
 static gboolean
 get_sim_file_id_and_path_with_separator (const gchar *file_path_str,
                                          guint16 *file_id,
@@ -1091,6 +1608,13 @@ get_sim_file_id_and_path_with_separator (const gchar *file_path_str,
     return TRUE;
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT
+        * HAVE_QMI_MESSAGE_UIM_READ_RECORD
+        * HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES */
+
+#if defined HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT || \
+    defined HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES
+
 static gboolean
 get_sim_file_id_and_path (const gchar *file_path_str,
                           guint16 *file_id,
@@ -1099,6 +1623,11 @@ get_sim_file_id_and_path (const gchar *file_path_str,
     return get_sim_file_id_and_path_with_separator (file_path_str, file_id, file_path, ",");
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT
+        * HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES */
+
+#if defined HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT
+
 static void
 read_transparent_ready (QmiClientUim *client,
                         GAsyncResult *res)
@@ -1201,6 +1730,10 @@ read_transparent_build_input (const gchar *file_path_str)
     return input;
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT */
+
+#if defined HAVE_QMI_MESSAGE_UIM_READ_RECORD
+
 static void
 read_record_ready (QmiClientUim *client,
                    GAsyncResult *res)
@@ -1297,17 +1830,31 @@ set_read_record_properties_handle (const gchar *key,
     }
 
     if (g_ascii_strcasecmp (key, "file") == 0) {
-        props->file = strdup(value);
+        props->file = g_strdup (value);
         return TRUE;
     }
 
     if (g_ascii_strcasecmp (key, "record-number") == 0) {
-        props->record_number = (guint16) atoi(value);
+        guint aux;
+
+        if (!qmicli_read_uint_from_string (value, &aux) || (aux > G_MAXUINT16)) {
+            g_set_error (error, QMI_CORE_ERROR, QMI_CORE_ERROR_FAILED,
+                         "failed reading key 'record-number' as 16bit value");
+            return FALSE;
+        }
+        props->record_number = (guint16) aux;
         return TRUE;
     }
 
     if (g_ascii_strcasecmp (key, "record-length") == 0) {
-        props->record_length = (guint16) atoi(value);
+        guint aux;
+
+        if (!qmicli_read_uint_from_string (value, &aux) || (aux > G_MAXUINT16)) {
+            g_set_error (error, QMI_CORE_ERROR, QMI_CORE_ERROR_FAILED,
+                         "failed reading key 'record-length' as 16bit value");
+            return FALSE;
+        }
+        props->record_length = (guint16) aux;
         return TRUE;
     }
 
@@ -1375,6 +1922,10 @@ out:
     return input;
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_READ_RECORD */
+
+#if defined HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES
+
 static void
 get_file_attributes_ready (QmiClientUim *client,
                            GAsyncResult *res,
@@ -1547,18 +2098,20 @@ get_file_attributes_build_input (const gchar *file_path_str)
     return input;
 }
 
+#endif /* HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES */
+
 void
 qmicli_uim_run (QmiDevice *device,
                 QmiClientUim *client,
                 GCancellable *cancellable)
 {
     /* Initialize context */
-    ctx = g_slice_new (Context);
+    ctx = g_slice_new0 (Context);
     ctx->device = g_object_ref (device);
     ctx->client = g_object_ref (client);
     ctx->cancellable = g_object_ref (cancellable);
 
-    /* Set PIN protection */
+#if defined HAVE_QMI_MESSAGE_UIM_SET_PIN_PROTECTION
     if (set_pin_protection_str) {
         QmiMessageUimSetPinProtectionInput *input;
 
@@ -1577,8 +2130,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_set_pin_protection_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to verify PIN? */
+#if defined HAVE_QMI_MESSAGE_UIM_VERIFY_PIN
     if (verify_pin_str) {
         QmiMessageUimVerifyPinInput *input;
 
@@ -1597,8 +2151,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_verify_pin_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to unblock PIN? */
+#if defined HAVE_QMI_MESSAGE_UIM_UNBLOCK_PIN
     if (unblock_pin_str) {
         QmiMessageUimUnblockPinInput *input;
 
@@ -1617,8 +2172,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_unblock_pin_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to change the PIN? */
+#if defined HAVE_QMI_MESSAGE_UIM_CHANGE_PIN
     if (change_pin_str) {
         QmiMessageUimChangePinInput *input;
 
@@ -1637,8 +2193,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_change_pin_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to read a transparent file? */
+#if defined HAVE_QMI_MESSAGE_UIM_READ_TRANSPARENT
     if (read_transparent_str) {
         QmiMessageUimReadTransparentInput *input;
 
@@ -1659,8 +2216,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_read_transparent_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to read a transparent file? */
+#if defined HAVE_QMI_MESSAGE_UIM_READ_RECORD
     if (read_record_str) {
         QmiMessageUimReadRecordInput *input;
 
@@ -1681,8 +2239,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_read_record_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to get file attributes? */
+#if defined HAVE_QMI_MESSAGE_UIM_GET_FILE_ATTRIBUTES
     if (get_file_attributes_str) {
         QmiMessageUimGetFileAttributesInput *input;
 
@@ -1703,8 +2262,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_get_file_attributes_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to read card status? */
+#if defined HAVE_QMI_MESSAGE_UIM_GET_CARD_STATUS
     if (get_card_status_flag) {
         g_debug ("Asynchronously getting card status...");
         qmi_client_uim_get_card_status (ctx->client,
@@ -1715,8 +2275,9 @@ qmicli_uim_run (QmiDevice *device,
                                         NULL);
         return;
     }
+#endif
 
-    /* Request to list supported messages? */
+#if defined HAVE_QMI_MESSAGE_UIM_GET_SUPPORTED_MESSAGES
     if (get_supported_messages_flag) {
         g_debug ("Asynchronously getting supported UIM messages...");
         qmi_client_uim_get_supported_messages (ctx->client,
@@ -1727,8 +2288,9 @@ qmicli_uim_run (QmiDevice *device,
                                                NULL);
         return;
     }
+#endif
 
-    /* Request to power on SIM card? */
+#if defined HAVE_QMI_MESSAGE_UIM_POWER_ON_SIM
     if (sim_power_on_str) {
         QmiMessageUimPowerOnSimInput *input;
 
@@ -1748,8 +2310,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_power_on_sim_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to power off SIM card? */
+#if defined HAVE_QMI_MESSAGE_UIM_POWER_OFF_SIM
     if (sim_power_off_str) {
         QmiMessageUimPowerOffSimInput *input;
 
@@ -1769,8 +2332,9 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_power_off_sim_input_unref (input);
         return;
     }
+#endif
 
-    /* Request to change provisioning session? */
+#if defined HAVE_QMI_MESSAGE_UIM_CHANGE_PROVISIONING_SESSION
     if (change_provisioning_session_str) {
         QmiMessageUimChangeProvisioningSessionInput *input;
 
@@ -1790,7 +2354,55 @@ qmicli_uim_run (QmiDevice *device,
         qmi_message_uim_change_provisioning_session_input_unref (input);
         return;
     }
+#endif
 
+#if defined HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS
+    /* Request to get slot status? */
+    if (get_slot_status_flag) {
+        g_debug ("Asynchronously getting slot status...");
+        qmi_client_uim_get_slot_status (ctx->client,
+                                        NULL,
+                                        10,
+                                        ctx->cancellable,
+                                        (GAsyncReadyCallback)get_slot_status_ready,
+                                        NULL);
+        return;
+    }
+#endif
+
+#if defined HAVE_QMI_MESSAGE_UIM_SWITCH_SLOT && defined HAVE_QMI_MESSAGE_UIM_GET_SLOT_STATUS
+    /* Request to change active slot? */
+    if (switch_slot_str) {
+        guint physical_slot;
+
+        if (!qmicli_read_uint_from_string (switch_slot_str, &physical_slot) ||
+            (physical_slot < 1) || (physical_slot > G_MAXUINT8)) {
+            g_printerr ("error: invalid slot number\n");
+            return;
+        }
+
+        g_debug ("Asynchronously switching active slot");
+
+        qmi_client_uim_get_slot_status (ctx->client,
+                                        NULL,
+                                        10,
+                                        ctx->cancellable,
+                                        (GAsyncReadyCallback)get_active_logical_slot_ready,
+                                        GUINT_TO_POINTER (physical_slot));
+        return;
+    }
+#endif
+
+#if defined HAVE_QMI_INDICATION_UIM_SLOT_STATUS
+    /* Watch for slot status changes? */
+    if (monitor_slot_status_flag) {
+        g_debug ("Listening for slot status changes...");
+        register_physical_slot_status_events ();
+        return;
+    }
+#endif
+
+#if defined HAVE_QMI_MESSAGE_UIM_RESET
     /* Request to reset UIM service? */
     if (reset_flag) {
         g_debug ("Asynchronously resetting UIM service...");
@@ -1802,6 +2414,7 @@ qmicli_uim_run (QmiDevice *device,
                               NULL);
         return;
     }
+#endif
 
     /* Just client allocate/release? */
     if (noop_flag) {
@@ -1811,3 +2424,5 @@ qmicli_uim_run (QmiDevice *device,
 
     g_warn_if_reached ();
 }
+
+#endif /* HAVE_QMI_SERVICE_UIM */
